/* -*-Mode:C; tab-width:4; indent-tabs-mode:t; c-file-style:"stroustrup";-*- */
#include "task.h"
#include "progect.h"
#include "progectdb.h"
#include "progectRsc.h"
#include "xb.h"
#include "projlistMODEL.h"


/****************************************************************************
 * Name : AddRecord
 * Desc : add a record to the database
 * In   : database, record to add
 * Out  : new index
 * Auth : lb, 27.07.2000
 * Mod  : lb, 03.08.2000
 * 			support to insert as a brother of selected task
 * Mod  : lb, 22.08.2000
 * 			added field for note
 * Mod  : rw, 30.12.2000
 * 			fixed bug in father inheritance, added category inheritance
 * Rem  : doesn't handle err codes
 ***************************************************************************/
Err AddRecord(DmOpenRef dbP, TaskTypePtr task, UInt16 *index)
{
	MemHandle h;
	TaskRecordType *p;
	UInt32 size = TaskRecordTypeSize;
	Err Err = errNone;
	UInt16 realIndex;
	// could be the clipboard
	UInt16 numRec = DmNumRecords(dbP);// - calcNumDeleted(dbP); 

	// pack this in a newly allocated record

	// calc the needed size
	// the two \0 are already in TaskRecordTypeSize
	size += (StrLen(task->description) + StrLen(task->note));

	// get a handle on a newly created record
	// new one always comes after actual
	if (numRec == 0)
	{
		realIndex = 0;
	}
	else if (numRec > 1 && TaskGetHasChild(dbP, gParentTask))
	{
		realIndex = TaskGetNextRelativeIndex(dbP, *index); // place of new one
	}
	else
	{
		realIndex = gParentTask + 1;
	}

	if (!realIndex)
	{
		realIndex = numRec;
	}
	h = DmNewRecord(dbP, &realIndex, size);

	task->attr.allBits = 0;
	task->attr.bits.opened = 1;
	task->dueDate.year = 0;
	task->dueDate.month = 0;
	task->dueDate.day = 0;

	// if index = 0, this is the first one
	// it's not a real task !!!
	if (realIndex == 0)
	{
		task->attr.bits.hasChild = 0;
	}
	else
	{
		// retrieve info about user of previous index
		// this sets : level, next
		if (*index == 0 || !TaskGetHasChild(dbP, gParentTask))
		{
			// this is a first child
			task->attr.bits.level = gRefLevel + 1;
			TaskSetHasChild(dbP, gParentTask, true);
		}
		else
		{
			task->attr.bits.level = TaskGetLevel(dbP, *index);
			task->attr.bits.hasPrev = 1;
			task->attr.bits.hasNext = TaskGetHasNext(dbP, *index);

			// update hasNext field in previous
			TaskSetHasNext(dbP, *index, true);

			if (gProjectPrefs.useFatherStatus)
			{
				if ((task->attr.bits.level) != FIRST_LEVEL)
				{
					UInt16 father = TaskGetFatherIndex(dbP, *index);
					task->dueDate = TaskGetDueDate(dbP, father);
					task->priority = TaskGetPriority(dbP, father);
				}
			}
			
			// inherit the default category
			if (gProjectPrefs.useFatherStatus)
			{
				if ((task->attr.bits.level) != FIRST_LEVEL)
				{
					UInt16 father = TaskGetFatherIndex(dbP, *index);
					UInt8 category;

					// get the father's category
					category = TaskGetCategory(gdbP, father);

					// set the default for the child
					TaskSetCategory(gdbP, realIndex, category);
				}
			}

		}
	}

	// get a pointer on the record
	p = MemHandleLock(h);

	// write the data
	DmWrite(p, 0, task, OffsetOf(TaskRecordType, description));
	DmStrCopy(p, OffsetOf(TaskRecordType, description), task->description);
	DmStrCopy(p, OffsetOf(TaskRecordType, description) + StrLen(task->description) + 1, task->note);

	// unlock the pointer
	MemHandleUnlock(h);

	// unlock the record
	DmReleaseRecord(dbP, realIndex, true);

	*index = realIndex;

//	DEBUG1("Added desc : ");
//	DEBUG1(task->description);
//	DEBUG1("Added note : ");
//	DEBUG1(task->note);
//	DEBUGVAL("len desc : ", StrLen(task->description));
//	DEBUGVAL("len note : ", StrLen(task->note));

	return Err;
} // static Err AddRecord(DmOpenRef dbP, TaskTypePtr task, UInt16 *index)



/****************************************************************************
 * Name : AddRecordSub
 * Desc : add a record to the database, as subtask of index
 * In   : database, record to add, index
 * Out  : new index
 * Auth : lb, 04.08.2000
 * Rem  : doesn't handle err codes
 * Mod  : lb, 13.08.2000
 * 			fix bug, can now create more subtasks. Subtasks always created
 * 			as last child
 * Mod  : lb, 22.08.2000
 * 			added field for note
 * Mod  : rw, 30.12.2000
 * 			fixed bug in father inheritance, added category inheritance
 ***************************************************************************/
Err AddRecordSub(DmOpenRef dbP, TaskTypePtr task, UInt16 *index)
{
	MemHandle h;
	TaskRecordType *p;
	UInt32 size = TaskRecordTypeSize;
	Err Err = errNone;
	UInt16 realIndex;

	// pack this in a newly allocated record

	// calc the needed size
	size += (StrLen(task->description) + StrLen(task->note));

	// get a handle on a newly created record
	// new one always comes after actual
	realIndex = TaskGetNextRelativeIndex(dbP, *index); // place of new one
	if (!realIndex)
	{
		realIndex = PgNumRecords(dbP);
	}
	h = DmNewRecord(dbP, &realIndex, size);

	// retrieve info about user of previous index
	// this sets : level, next
	task->attr.allBits = 0;
	task->attr.bits.level = TaskGetLevel(dbP, *index) + 1;
	task->attr.bits.opened = 1;
	task->attr.bits.hasPrev = TaskGetHasChild(dbP, *index);

	//task->priority = 6;
	task->dueDate.year = 0;
	task->dueDate.month = 0;
	task->dueDate.day = 0;

	if (gProjectPrefs.useFatherStatus)
	{
		if ((task->attr.bits.level) != FIRST_LEVEL)
		{
			task->dueDate = TaskGetDueDate(dbP, *index);
			task->priority = TaskGetPriority(dbP, *index);
		}
	}

	// inherit the default category
	if (gProjectPrefs.useFatherStatus)
	{
		if ((task->attr.bits.level) != FIRST_LEVEL)
		{
			UInt8 category;

			// get the father's category
			category = TaskGetCategory(gdbP, *index);

			// set the default for the child
			TaskSetCategory(gdbP, realIndex, category);
		}
	}

	// if father had children already
	if (TaskGetHasChild(dbP, *index))
	{
		// search prev task
		UInt16 n = realIndex - 1;
		while (TaskGetLevel(dbP, n) > task->attr.bits.level)
			n--;

		// old last child has now a next
		TaskSetHasNext(dbP, n, true);
	}

	// update hasChild field in father
	TaskSetHasChild(dbP, *index, true);

	// get a pointer on the record
	p = MemHandleLock(h);

	// write the data
	DmWrite(p, 0, task, OffsetOf(TaskRecordType, description));
	DmStrCopy(p, OffsetOf(TaskRecordType, description), task->description);
	DmStrCopy(p, OffsetOf(TaskRecordType, description) + StrLen(task->description) + 1, task->note);

	// unlock the pointer
	MemHandleUnlock(h);

	// unlock the record
	DmReleaseRecord(dbP, realIndex, true);

	*index = realIndex;

	return Err;
} // static Err AddRecordSub(DmOpenRef dbP, TaskTypePtr task, UInt16 *index)



/****************************************************************************
 * Name : CreateDB
 * Desc : create a database
 * In   : 
 * 			-> name without prepanding !
 * Out  : -
 * Auth : lb, 26.07.2000
 * Rem  : doesn't handle err codes
 * Mod  : lb, 17.08.2000
 * 			- prepand "lbPG-" to the dbname
 ***************************************************************************/
Err CreateDB(const char *nameP)
{
	UInt16 cardNo = 0;
	Char dbname[dmDBNameLength];
	Err Err = errNone;
	LocalID dbID;
	TaskType task0;
	UInt16 index = 0;
	UInt16 dbversion = DBVERSION;
	UInt16 attr;

	SafeCopy(dbname,nameP,dmDBNameLength);
	Prepend(dbname);

	dbID = DmFindDatabase(cardNo, dbname);

	// if database doesn't exist, create it
	if (!dbID)
	{
		Err = DmCreateDatabase(cardNo, dbname, CREATOR, 'DATA', false);
		dbID = DmFindDatabase(cardNo, dbname);
		if (!dbID)
		{
			return Err;
		}
		// make sure db is backed up
		DmDatabaseInfo(0, dbID, NULL, &attr, NULL, NULL, NULL, 
			NULL, NULL, NULL, NULL, NULL, NULL);
		attr = attr | dmHdrAttrBackup;

		// set database format version & attributes
		DmSetDatabaseInfo(cardNo, dbID, NULL, &attr, &dbversion, NULL, NULL,
			NULL, NULL, NULL, NULL, NULL, NULL);

		// Need to create the initial record for some reason
		// OpenDB takes a name without prepend
		OpenDB(cardNo,nameP, &gdbP);
		task0 = gEmptyTask;
		AddRecord(gdbP, &task0, &index);
		CloseDB(gdbP);
		gdbP = NULL;
		ProjectListMODELInit();
		ProjectListModelAddProject(cardNo,nameP);
		ProjectListMODELCleanUp();
	} else
	{
		Err = dmErrAlreadyExists;
	}

	return Err;
} // static Err CreateDB(const char *nameP)


/****************************************************************************
 * Name : RemoveDB
 * Desc : remove a database
 * In   : name of database to remove
 * Out  : Err code
 * Auth : lb, 27.07.2000
 * Mod  : lb, 17.08.2000
 * 			prepending support
 * 			control of CREATOR ID
 ***************************************************************************/
Err RemoveDB(const char *nameP)
{
	UInt16 cardNo = 0;
	Err Err = errNone;
	LocalID dbID;
	Char dbname[dmDBNameLength];
	UInt32 creatorID;

	SafeCopy(dbname,nameP,dmDBNameLength);
	Prepend(dbname);

	dbID = DmFindDatabase(cardNo, dbname);

	// if database exists, delete it
	if (dbID)
	{
		DmDatabaseInfo(cardNo, dbID, NULL, NULL, NULL, NULL, NULL, 
			NULL, NULL, NULL, NULL, NULL, &creatorID);
		if (creatorID == CREATOR)
		{
			Err = DmDeleteDatabase(cardNo, dbID);
		}

		//
		// Update project list model
		//
		if (errNone == Err)
		{
			ProjectListMODELInit();
			ProjectListModelDeleteProject(cardNo,nameP);
			ProjectListMODELCleanUp();
		}
	}

	return Err;
} // static Err RemoveDB(const char *nameP)



/****************************************************************************
 * Name : PgNumRecords
 * Desc : Return the number of non-deleted records in the database
 * Prm  : 
 * 			-> dbP, database reference
 * Out  : num of non-deleted records
 * Rem  : gNumDeleted MUST be uptodate
 * Auth : lb, 2001-09-05
 * Mod  : lb, 2002-01-21
 * 			gNumDeleted not used anymore
 ***************************************************************************/
UInt16 PgNumRecords(DmOpenRef dbP)
{
	return DmNumRecords(dbP);
	//return DmNumRecords(dbP) - gNumDeleted;
} // UInt16 PgNumRecords(DmOpenRef dbP)



/****************************************************************************
 * Name : calcNumDeleted
 * Desc : calc and init gNumDeleted, the number of deleted records in dbP
 * Prm  : 
 * 			-> dbP, database reference
 * Out  : number of deleted records
 * Auth : lb, 2001-09-04
 * Mod  : lb, 2002-01-21
 * 			Not used anymore
 ***************************************************************************/
UInt16 calcNumDeleted(DmOpenRef dbP)
{
//	Int32 i = (Int32)DmNumRecords(dbP) - 1;
//	UInt16 attr;
//	Int32 num = -1;
//	DBGMSG((DBB, "In calcNumDeleted"));
//	attr = dmRecAttrDelete;
//	while ((attr & dmRecAttrDelete) && (i > 0))
//	{
//		num++;
//		DmRecordInfo(dbP, i--, &attr, NULL, NULL);
//	}
//	if (num == -1)
//		num = 0;
//	DBGMSG((DBB, "Num deleted : %lu", num));
//	return (UInt16)num;
	return 0;
} // UInt16 calcNumDeleted(DmOpenRef dbP)



/****************************************************************************
 * Name : OpenDB
 * Desc : open a database by name
 * Prm  : 
 * 			->  name of the db to open without prepending
 * 			<-> database reference
 * Out  : Err code
 * Auth : lb, 27.07.2000
 * Mod  : lb, 14.08.2000
 * 			to save app current state
 * 		  lb, 17.08.2000
 * 		  	try to prepend
 * 		  lb, 24.08.2000
 * 		  	fix without prepending mode that caused a crash after two
 * 		  	come backs
 ***************************************************************************/
Err OpenDB(const UInt16 cardNo, const char* nameP, DmOpenRef *dbP)
{
	UInt16 mode = dmModeReadWrite;
	UInt16 dbversion;
	Err Err = 0;
	Char dbname[dmDBNameLength];

	DBGMSG((DBB, "In OpenDB"));

	SafeCopy(dbname,nameP,dmDBNameLength);
	Prepend(dbname);

	// find the database
	gdbID = DmFindDatabase(cardNo, dbname);
	if (!gdbID)
	{
		// try without prepending (for databases before V0.10)
		// find the database
		SafeCopy(dbname,nameP,dmDBNameLength);
		gdbID = DmFindDatabase(cardNo, dbname);

		// if found
		if (!gdbID)
		{
			return DmGetLastErr();
		}
	}

	// Check version number
	DmDatabaseInfo(cardNo, gdbID, NULL, NULL, &dbversion, NULL, NULL, NULL,
		NULL, NULL, NULL, NULL, NULL);
	if (dbversion > DBVERSION)
	{
		Char version[32];
		Char* p;
		StrIToA(version, dbversion/100);
		p = version + StrLen(version);
		*p = '.';
		p++;
		StrIToA(p, dbversion % 100);
		
		DEBUG2("The project you are trying to open needs version ^1 of "
			"Progect. Please upgrade at sf.net/projects/progect", version);
		return 1;
	}

	// open it
	*dbP = DmOpenDatabase(cardNo, gdbID, mode);

	// if it failed
	if (!*dbP)
	{
		// get error code
		Err = DmGetLastErr();
	}
	else
	{
		SafeCopy(gCurrentPrefs.openDBName, dbname,dmDBNameLength);
		gCurrentPrefs.openDB = true;
		LoadDBPrefs(dbP);
		//gNumDeleted = calcNumDeleted(*dbP);
		//DBGMSG((DBB, "num deleted at opening : %hu", gNumDeleted));
		if (gProjectPrefs.autoSyncToDo)
		{
			TaskSyncAll();
			DBGMSG((DBB, "After task sync all"));
		}
		// save/restore the pointer have no meaning,
		// and it will be crash program.
		//   NOTE: fix BUG#115440
		//             "Newly created task/note has garbage text in it"
		MemMove((MemPtr)&gEmptyTask,
				(MemPtr)&gProjectPrefs.taskDefaults,
				(UInt8*)&(gEmptyTask.description) - (UInt8*)&gEmptyTask);
//		gEmptyTask = gProjectPrefs.taskDefaults;

               // make sure a corrupt data structure does not crash program
               if (gEmptyTask.format.bits.itemType >= invalidType)
			gEmptyTask.format.bits.itemType=progressType;

		// needed by 0.23 to get rid of the deleted records
	}

	return Err;
} // static Err OpenDB(Uint16 cardNo,const char* nameP, DmOpenRef *dbP)


/****************************************************************************
 * Name : OpenDBByID
 * Desc : open a database by cardNo and dbID
 * Prm  : 
 * 			->  card no
 * 			->  database id
 * 			<-> database reference
 * Out  : Err code
 * Mod  : lb, 2001-12-03
 * 		fixed bug causing loss of categories when using history
 ***************************************************************************/
Err OpenDBByID(UInt16 cardNo, LocalID dbID, DmOpenRef* dbP)
{
	// Open 
	Char* name;
	DmOpenRef pDB = DmOpenDatabase(cardNo, dbID, dmModeReadWrite);
	if (! pDB)
		return DmGetLastErr();

	// close current opend DB
	if (*dbP)
		CloseDB(*dbP);

	// setting up global status.
	*dbP = pDB;
	gdbID = dbID;

	name = MemPtrNew(dmDBNameLength);
	DmDatabaseInfo(cardNo, dbID, name, NULL, NULL, NULL, NULL, NULL,
				   NULL, NULL, NULL, NULL, NULL);
	SafeCopy(gCurrentPrefs.openDBName, name,dmDBNameLength);
	MemPtrFree(name);
	gCurrentPrefs.openDB = true;
	LoadDBPrefs(dbP);
	//gNumDeleted = calcNumDeleted(*dbP);
	if (gProjectPrefs.autoSyncToDo)
		TaskSyncAll();

	return 0;
} // Err OpenDBByID(UInt16 cardNo, LocalID dbID, DmOpenRef* dbP)



/****************************************************************************
 * Name : CloseDB
 * Desc : close the database
 * In   : pointer to the database to close
 * Out  : Err code
 * Auth : lb, 27.07.2000
 * Mod  : lb, 14.08.2000
 * 			to save app current state
 ***************************************************************************/
Err CloseDB(DmOpenRef dbP)
{
	Err Err;
	SaveDBPrefs(dbP);
	// close the database
	Err = DmCloseDatabase(dbP);
	gCurrentPrefs.openDB = false;
	gCurrentPrefs.topTask = -1;
	gdbID = 0;
	gdbP = NULL;
	return Err;
} // static Err CloseDB(DmOpenRef dbP)



/****************************************************************************
 * Name : RenameDB
 * Desc : rename a database
 * In   :
 * 			-> old name (with or without prepending)
 * 			-> new name (with or without prepending)
 * Out  : -
 * Auth : lb, 23.08.2000
 ***************************************************************************/
Err RenameDB(Char* oldName, Char* newName)
{
	UInt16 cardNo = 0;
	LocalID dbID;
	Err err;
	Char dbnameOld[dmDBNameLength];
	Char dbnameNew[dmDBNameLength];

	SafeCopy(dbnameOld, oldName,dmDBNameLength);
	SafeCopy(dbnameNew, newName,dmDBNameLength);
	Prepend(dbnameOld);
	Prepend(dbnameNew);

	// find the database
	if (!(dbID = DmFindDatabase(cardNo, dbnameOld)))
		return DmGetLastErr();

	err = DmSetDatabaseInfo(cardNo, dbID, dbnameNew, NULL, NULL, NULL, NULL, 
					NULL, NULL, NULL, NULL, NULL, NULL);
	//
	// Update the project list model
	//
	if (errNone == err)
	{
		ProjectListMODELInit();
		ProjectListModelRenameProject(cardNo,oldName,newName);
		ProjectListMODELCleanUp();
	}

	return err;
} // static Err RenameDB(Char* oldName, Char* newName)



/****************************************************************************
 * Name : SaveDBPrefs
 * Desc : save the database preferences
 * In   :
 * Out  : -
 * Auth : lb, 17.08.2000
 ***************************************************************************/
void SaveDBPrefs(DmOpenRef dbP)
{
	LocalID appInfoID;
	DBInfoType* p;
	MemHandle h;

	appInfoID = DmGetAppInfoID(dbP);

	// if there is no appInfoID, create it
	if (! appInfoID)
	{
		h = DmNewHandle(dbP, sizeof(DBInfoType));
		if (!h)
			return;
		appInfoID = MemHandleToLocalID(h);
		DmSetDatabaseInfo(0, gdbID, NULL, NULL, NULL, NULL, NULL, NULL, NULL, 
			&appInfoID, NULL, NULL, NULL);

		p = MemLocalIDToLockedPtr(appInfoID, 0);
		DmSet(p, 0, sizeof(DBInfoType), 0);
		CategoryInitialize(&p->appInfo, InitialCategory);
	}
	else
		p = MemLocalIDToLockedPtr(appInfoID, 0);

	// test the size
	// if it's not good, the next LoadDBPrefs will correct it
	if (MemPtrSize(p) == sizeof(DBInfoType))
	{
		// write the new prefs
		DmWrite(p, OffsetOf(DBInfoType, pref),
				&gProjectPrefs, sizeof(ProjectPrefsType));
	}
	// unlock
	MemPtrUnlock(p);

} // static void SaveDBPrefs(void)



/****************************************************************************
 * Name : LoadDBPrefs
 * Desc : load the database preferences
 * In   :	-> ptr to the DmOpenRef
 * Out  : false = no prefs
 * 		  true  = ok
 * 		  DmOpenRef is updated if the database has changed (conversion)
 * Auth : lb, 17.08.2000
 * Mod  : lb, 17.08.2000
 * 			- pass db, because when OpenDB calls, gdbp is not set !
 *        hcc, 05.05.2001
 *          - When resizing project preferences, preserve old prefs.
 ***************************************************************************/
Boolean LoadDBPrefs(DmOpenRef *dbP)
{
	DBInfoType *p;
	LocalID appInfoID;
	LocalID oldInfoID = 0;
	UInt32 len;

	DBGMSG((DBB, "In LoadDBPrefs"));
	appInfoID = DmGetAppInfoID(*dbP);

	if (appInfoID)
	{
		p = MemLocalIDToLockedPtr(appInfoID, 0);
		if (MemPtrSize(p) < 30 && *(UInt8*)p < 11)
		{
			DEBUG1("Old database version (<=0.11) conversion, code removed...");
			MemPtrUnlock(p);
			/* TODO: Finish removing this logic...
             * ConvertDB(11);*/
		}
		else if (p->pref.format < 23)
		{
			DEBUG1("Converting db to 0.23 format, please click OK and wait...");
			MemPtrUnlock(p);
			*dbP = ConvertDB(18);
			oldInfoID = DmGetAppInfoID(*dbP);
			appInfoID = 0;
		}
		else if (p->pref.format == 23)
		{
			// 0.23 database could have deleted records, we purge them since
			// they're not needed at all
			TaskRemoveDeletedRecords(*dbP);
		}
		else if (MemPtrSize(p) != sizeof(DBInfoType) )
		{
			MessageBox(StrPrefsHasBeenDeleted);
			oldInfoID = appInfoID;
			appInfoID = 0;
			MemPtrUnlock(p);
		}
	}

	if (! appInfoID)
	{
		// create new prefs
		MemHandle h = DmNewHandle(*dbP, sizeof(DBInfoType));
		if (! h)
			return false;
		DBGMSG((DBB,"need to reconstruct appinfoID"));
		appInfoID = MemHandleToLocalID(h);
		DmSetDatabaseInfo(0, gdbID, NULL, NULL, NULL, NULL, NULL, 
						  NULL, NULL, &appInfoID, NULL, NULL, NULL);
		// get a pointer
		p = MemLocalIDToLockedPtr(appInfoID, 0);
		
		// write the new prefs
		DmSet(p, 0, sizeof(DBInfoType), 0);

		// Init prefs to default. May be partially overwritten with
		// newer prefs below.
		DmWrite(p, OffsetOf(DBInfoType, pref),
				&gProjectPrefs, sizeof(ProjectPrefsType));
		CategoryInitialize(&p->appInfo, InitialCategory);

		if (oldInfoID)
		{   // copy original categories & preferences
			AppInfoType* pp = MemLocalIDToLockedPtr(oldInfoID, 0);
			DBGMSG((DBB,"Old appinfoID found, using it"));

			if (MemPtrSize(pp) > sizeof(AppInfoType) + 20)
			{
				DmWrite(p, 0, pp, sizeof(AppInfoType));

				// Calculate amount of old prefs to copy, based on
				// the smaller set of prefs.
				len = MemPtrSize(pp);
				if(len > MemPtrSize(p))
					len = MemPtrSize(p);
				len -= sizeof(AppInfoType);
				
				DmWrite(p, sizeof(AppInfoType),
						((char *)pp) + sizeof(AppInfoType), len);
			}
			MemPtrUnlock(pp);
		}
		// set new DB format (+ copy reserved field)
		DmWrite(p, 0, &gProjectPrefs, sizeof(UInt16));
	}
	// this produced an overlocked pointer => crash every 8th start
	//else
		//p = MemLocalIDToLockedPtr(appInfoID, 0);

	MemMove(&gProjectPrefs, &p->pref, sizeof(ProjectPrefsType));
	MemPtrUnlock(p);


	if (oldInfoID)
	{
		// ... ??? how to delete old ID ???
	}

	return true;
} // static Boolean LoadDBPrefs(DmOpenRef *dbP)



/****************************************************************************
 * Name : SaveAppPrefs
 * Desc : save the application preferences
 * In   :
 * 			-> saved : true for saved prefs, false for current prefs
 * Out  : -
 * Auth : lb, 14.08.2000
 * Mod  : lb, 17.08.2000
 * 			support for saved prefs
 ***************************************************************************/
void SaveAppPrefs(Boolean saved)
{
	PrefSetAppPreferences(CREATOR, 0, 1, 
		saved ? (void*)&gSavedPrefs : (void*)&gCurrentPrefs, 
		saved ? sizeof(SavedPrefsType) : sizeof(CurrentPrefsType), 
		saved);
} // static void SaveAppPrefs(Boolean saved)



/****************************************************************************
 * Name : LoadAppPrefs
 * Desc : load the application preferences
 * In   :
 * 			-> saved : true for saved prefs, false for current prefs
 * Out  : false = no prefs
 * 		  true  = ok
 * Auth : lb, 14.08.2000
 * Mod  : lb, 17.08.2000
 * 			support for saved prefs
 ***************************************************************************/
Boolean LoadAppPrefs(Boolean saved)
{
	UInt16 size = saved ? sizeof(SavedPrefsType) : sizeof(CurrentPrefsType);
	UInt16 status;

	status = PrefGetAppPreferences(CREATOR, 0, 
		saved ? (void*)&gSavedPrefs : (void*)&gCurrentPrefs, &size, saved);
	status = (size == 
		(saved ? sizeof(SavedPrefsType) : sizeof(CurrentPrefsType))) ? status : noPreferenceFound;
	return status == noPreferenceFound ? false : true;
} // static Boolean LoadAppPrefs(Boolean saved)



/****************************************************************************
 * Name : ConvertDB
 * Desc : convert an old DB to the actual format
 * In   :
 * 			-> old format
 * Out  :   <- new DmOpenRef
 * Auth : lb, 22.08.2000
 * Mod  : lb, 2001-10-16 : conversion 0.22 -> 0.23
 ***************************************************************************/
DmOpenRef ConvertDB(UInt8 dbformat)
{
	Char dbname[dmDBNameLength];
	UInt16 recNum;
	Char* tmpName = "lbPG-tmpConvertProcess";
	LocalID newID, oldID;
	DmOpenRef newDB = 0, oldDB;
	MemHandle hOld, hNew;
	UInt16 i, descLen;
	UInt16 fakeIndex;
	UInt32 size, xbSize;
	UInt16 cardNo = 0;
	TaskFormatType format;
	TaskAttrType attr;
	UInt8 completed;

	SafeCopy(dbname, gCurrentPrefs.openDBName,dmDBNameLength);
	switch (dbformat)
	{
		/* Starting with 0.23, removed support for version 0.11 conversion. */
		case 18: // old(0.18) -> 0.23
			/* For this conversion, the following things must be done :
			 * 		- replace changed bits of TaskAttrType
			 * 		- update task type
			 * 		- update progress status
			 * 		- move XB
			 */
			DBGMSG((DBB, "Closing database"));
			DmCloseDatabase(gdbP);
			DBGMSG((DBB, "Renaming database"));
			RenameDB(dbname, tmpName);

			// prepend for opening
			Prepend(dbname);

			// create, pass the name without prepending
			CreateDB(&dbname[5]);
			// open (prepended name)
			newID = DmFindDatabase(cardNo, dbname);

			// open the databases
			DBGMSG((DBB, "Searching old database"));
			oldID = DmFindDatabase(cardNo, tmpName);
			DBGMSG((DBB, "Opening new"));
			newDB = DmOpenDatabase(cardNo, newID, dmModeReadWrite);
			DBGMSG((DBB, "Opening old"));
			oldDB = DmOpenDatabase(cardNo, oldID, dmModeReadOnly);
			
			// get the number of records
			recNum = DmNumRecords(oldDB);// - calcNumDeleted(oldDB);
			DBGMSG((DBB, "%hu records found in old database", recNum));

			// convert each record (except task0 and the deleted ones)
			for (i = 1; i < recNum; i++)
			{
				TaskRecordType *pOld;
				TaskRecordType *pNew;

				// get the old one
				DBGMSG((DBB, "Getting old record %hu", i));
				hOld = DmQueryRecord(oldDB, i);
				pOld = MemHandleLock(hOld);
				fakeIndex = dmMaxRecordIndex;
				DBGMSG((DBB, "Desc : %s", &pOld->description));

				// calc the size of the new one
				descLen = StrLen(&pOld->description) + 1; // with \0
				size = TaskRecordTypeSize + StrLen(&pOld->description) +
					StrLen(&pOld->description + descLen); 
				DBGMSG((DBB, "Computed record size : %lu", size));

				// create a record
				hNew = DmNewRecord(newDB, &fakeIndex, size);
				
				// reget the pointer
				MemHandleUnlock(hOld);
				pOld = MemHandleLock(hOld);

				// convert attributes
				attr.level = pOld->attr.bits.level;
				attr.hasNext = pOld->attr.bits.hasNext;
				attr.hasChild = pOld->attr.bits.hasChild;
				attr.opened = pOld->attr.bits.opened;
				attr.hasPrev = pOld->attr.bits.hasPrev;
				pNew = MemHandleLock(hNew);

				// write
				// first, reset it (to have the note be \0)
				DmSet(pNew, 0, size, 0);

				// write attr, priority, dueDate
				DBGMSG((DBB, "Writing attr, priority, duedate"));
				DmWrite(pNew, OffsetOf(TaskRecordType, attr), &attr, 
					sizeof(UInt16));
				DmWrite(pNew, OffsetOf(TaskRecordType, priority), 
					&pOld->priority, sizeof(UInt8));
				DmWrite(pNew, OffsetOf(TaskRecordType, dueDate), 
					&pOld->dueDate, sizeof(DateType));

				// copy description
				DBGMSG((DBB, "Writing description"));
				DmStrCopy(pNew, OffsetOf(TaskRecordType, description), 
					&pOld->description);

				// copy note
				DBGMSG((DBB, "Writing note"));
				DmStrCopy(pNew, OffsetOf(TaskRecordType, description) + descLen,
					&pOld->description + descLen);


				// special cases, completed and format
				format = pOld->format.bits;
				completed = pOld->completed;
				DBGMSG((DBB, "Completed = %hu", completed));
				if (pOld->completed < CNV_ACTION)
				{
					format.itemType = progressType;
					DBGMSG((DBB, "was a progress"));
				}
				else if (pOld->completed == CNV_ACTION_NO)
				{
					format.itemType = actionType;
					completed = 0;
					DBGMSG((DBB, "was an undone action"));
				}
				else if (pOld->completed == CNV_ACTION_OK)
				{
					format.itemType = actionType;
					completed = 10;
					DBGMSG((DBB, "was a done action"));
				}
				else if (pOld->completed == CNV_INFORMATIVE)
				{
					format.itemType = informativeType;
					completed = 0;
					DBGMSG((DBB, "was info"));
				}
				else if (pOld->completed >= CNV_NUMERIC)
				{
					format.itemType = numericType;
					completed = completed - CNV_NUMERIC;
					DBGMSG((DBB, "was numeric"));
				}
				DBGMSG((DBB, "writing completed and format"));
				DmWrite(pNew, OffsetOf(TaskRecordType, completed), 
					&completed, sizeof(UInt8));
				DmWrite(pNew, OffsetOf(TaskRecordType, format), &format, 
					sizeof(UInt16));

				MemHandleUnlock(hOld);
				MemHandleUnlock(hNew);
				DmReleaseRecord(newDB, fakeIndex, true);
				DBGMSG((DBB, "record released, getting extrablock"));

				xbSize = TaskGetExtraBlockV022(oldDB, i, NULL);
				DBGMSG((DBB, "xb size = %lu", xbSize));
				if (xbSize > 8) // one data chunk, one null chunk
				{
					MemPtr xb;
					xb = MemPtrNew(xbSize);
					DBGMSG((DBB, "getting xb"));
					TaskGetExtraBlockV022(oldDB, i, xb);
					DBGMSG((DBB, "setting xb"));
					TaskSetExtraBlock(newDB, i, xb, xbSize);
				}
				DBGMSG((DBB, "-------------------------"));
			}
			DmCloseDatabase(oldDB);
			DBGMSG((DBB, "deleting old db"));
			DmDeleteDatabase(cardNo, oldID);
			gdbID = newID;
			break;
		default:
			break;
	}
	return newDB;
} // static void ConvertDB(UInt8 format)



/****************************************************************************
 * Name : Prepend
 * Desc : check if a name is prepended, if not, prepend it
 * In   :
 * 			<-> name
 * Out  : -
 * Auth : lb, 26.08.2000
 ***************************************************************************/
void Prepend(Char* name)
{
	Char tmp[dmDBNameLength];
	// if the name is not prepended
	if (StrStr(name, gPrepend) != name)
	{
		// prepend the name
		SafeCopy(tmp, gPrepend,dmDBNameLength);
		StrNCat(tmp, name,dmDBNameLength);
		StrCopy(name, tmp); // Safe to do a StrCopy, we know src size
	}
} // void Prepend(Char* name)



/****************************************************************************
 * Name : DuplicateDB
 * Desc : duplicate a project
 * In   :
 * 			-> source (with or without prepending)
 * 			-> dest   ( "         "      "       )
 * Out  : -
 * Auth : lb, 27.08.2000
 ***************************************************************************/
Err DuplicateDB(const Char* srcIn, const Char* dbnameIn)
{
	Char src[dmDBNameLength];
	Char dbname[dmDBNameLength];
	UInt16 recNum;
	LocalID newID, oldID;
	DmOpenRef newDB, oldDB;
	MemHandle hOld, hNew;
	TaskRecordType *pOld, *pNew;
	UInt16 i;
	UInt16 index;
	UInt16 cardNo = 0;
	UInt16 oldAttr, newAttr;
	LocalID appInfoID;
	Err err;
	MemPtr oldP, newP;

	// make duplicates so we can modify
	SafeCopy(src, srcIn,dmDBNameLength);
	SafeCopy(dbname, dbnameIn,dmDBNameLength);

	// prepend _always_
	Prepend(dbname);
	
	// create, pass the name without the prepending
	err = CreateDB(&dbname[5]);
	// open (prepended name)
	if (!err)
		newID = DmFindDatabase(cardNo, dbname);
	else
		return err;

	// prepend (if needed)
	Prepend(src);

	// open the databases
	oldID = DmFindDatabase(cardNo, src);
	newDB = DmOpenDatabase(cardNo, newID, dmModeReadWrite);
	oldDB = DmOpenDatabase(cardNo, oldID, dmModeReadOnly);
	
	// get the number of records
	recNum = DmNumRecords(oldDB);// - calcNumDeleted(oldDB);

	// copy each record (except task0)
	for (i = 1; i < recNum; i++)
	{
		index = dmMaxRecordIndex;
		hOld = DmQueryRecord(oldDB, i);
		pOld = MemHandleLock(hOld);
		hNew = DmNewRecord(newDB, &index, MemPtrSize(pOld));
		if (!hNew)
		{
			MessageBox(StrNoMemorySpaces);
			break;
		}
		pNew = MemHandleLock(hNew);
		DmWrite(pNew, 0, pOld, MemPtrSize(pOld));
		MemHandleUnlock(hNew);
		MemHandleUnlock(hOld);
		DmReleaseRecord(newDB, index, true);
		// get category
		DmRecordInfo(oldDB, i, &oldAttr, NULL, NULL);
		DmRecordInfo(newDB, i, &newAttr, NULL, NULL);
		newAttr &= ~dmRecAttrCategoryMask; // remove category info
		oldAttr &= dmRecAttrCategoryMask; // keep only category info
		newAttr |= oldAttr; // get old category
		// set it to the new record
		DmSetRecordInfo(newDB, index, &newAttr, NULL);
	}

	// copy appInfo block
	appInfoID = DmGetAppInfoID(oldDB);
	oldP = MemLocalIDToLockedPtr(appInfoID, 0);

	hNew = DmNewHandle(newDB, MemPtrSize(oldP));
	if (!hNew)
	{
		MessageBox(StrNoMemorySpaces);
	}
	else
	{
		appInfoID = MemHandleToLocalID(hNew);
		DmSetDatabaseInfo(0, newID, NULL, NULL, NULL, NULL, NULL, NULL, NULL, 
			&appInfoID, NULL, NULL, NULL);

		newP = MemLocalIDToLockedPtr(appInfoID, 0);
		DmWrite(newP, 0, oldP, MemPtrSize(oldP));

		MemPtrUnlock(oldP);
		MemPtrUnlock(newP);
	}

	DmCloseDatabase(oldDB);
	DmCloseDatabase(newDB);

	return errNone;
} // void DuplicateDB(Char* src, Char* dbname)



/****************************************************************************
 * Name : OpenClipboard
 * Desc : open the clipboard, create it if it doesn't exist
 * In   :
 * Out  : -
 * Auth : lb, 04.09.2000
 ***************************************************************************/
pgErr OpenClipboard(void)
{
	UInt16 cardNo = 0;
	Err err = errNone;
	LocalID dbID;
	char dbname[] = "lbPG-Clipboard";
	UInt16 index = 0;

	dbID = DmFindDatabase(cardNo, dbname);

	// if database doesn't exist, create it
	if (!dbID)
	{
		err = DmCreateDatabase(cardNo, dbname, CREATOR, 'CLIP', false);
		if (err)
		{
			DEBUG1("Cannot create a clipboard database.");
			return pgError;
		}
		dbID = DmFindDatabase(cardNo, dbname);
	}

	gClip = DmOpenDatabase(cardNo, dbID, dmModeReadWrite);
	if (DmNumRecords(gClip) == 0)
	{
		DEBUG1("Adding task 0");
		AddRecord(gClip, &gEmptyTask, &index);
	}
	return pgOK;
} // pgErr OpenClipboard(void)



/****************************************************************************
 * Name : CloseClipboard
 * Desc : close the clipboard
 * In   :
 * Out  : -
 * Auth : lb, 04.09.2000
 ***************************************************************************/
pgErr CloseClipboard(void)
{
	if (gClip)
	{
		DmCloseDatabase(gClip);
		gClip = NULL;
	}

	return pgOK;
} // pgErr CloseClipboard(void)



/****************************************************************************
 * Name : NewClipboard
 * Desc : close the clipboard, remove it, create a new one, open it
 * In   :
 * Out  : -
 * Auth : lb, 05.09.2000
 ***************************************************************************/
pgErr NewClipboard(void)
{
	UInt16 cardNo = 0;
	Err err = errNone;
	LocalID dbID;
	char dbname[] = "lbPG-Clipboard";
	UInt16 index = 0;

	CloseClipboard();

	dbID = DmFindDatabase(cardNo, dbname);

	// if database exists, remove it
	if (dbID)
	{
		DmDeleteDatabase(cardNo, dbID);
	}

	err = DmCreateDatabase(cardNo, dbname, CREATOR, 'CLIP', false);
	if (err)
		return pgError;

	dbID = DmFindDatabase(cardNo, dbname);

	gClip = DmOpenDatabase(cardNo, dbID, dmModeReadWrite);
	AddRecord(gClip, &gEmptyTask, &index);

	return pgOK;
		
} // pgErr NewClipboard(void)

/****************************************************************************
 * Name : SafeCopy
 * Desc : Do all the safe copy stuff, because StrNCopy is still stupid.
 * In   : dest, src (overwritten), size_of_buffer
 * Out  : -
 * Auth : rp, 03.08.2003
****************************************************************************/
void SafeCopy(Char* dest,const Char *src, UInt16 size)
{
	// +++ FIX THIS +++
	// Need to check for a size of zero, since that's invalid
	// +++ FIX THIS +++

	// StrnCopy can only take a UInt16 for size
	StrNCopy(dest,src,size-1);
	dest[size-1]=chrNull;
} // void SafeCopy(Char* dest,const Char *src, UInt16 size)
